import urllib3

from requezts import _HTTPConnection


class _HTTPConnectionPool(urllib3.connectionpool.HTTPConnectionPool):
    """HTTPConnectionPool using libzt sockets

    Internal object to requezts. Should not be called by user code.
    """
    ConnectionCls = _HTTPConnection



from ._socket import _socket, _SocketIO
from ._connection import _HTTPConnection
from ._connectionpool import _HTTPConnectionPool
from ._poolmanager import _PoolManager
from ._adapters import _HTTPAdapter
from .session import Session



from typing import Optional

import urllib3
import libzt

from requezts import _socket


# noinspection PyUnresolvedReferences,PyProtectedMember
class _HTTPConnection(urllib3.connection._HTTPConnection):
    """HTTPConnection using libzt sockets

    Internal object to requezts. Should not be called by user code.
    """

    def _new_conn(self):
        """Establish a new socket/connection"""

        host = self._dns_host
        if host.startswith("["):
            host = host.strip("[]")
        try:
            host.encode("idna")
        except UnicodeError:
            raise urllib3.exceptions.LocationParseError(f"'{host}', label empty or too long") from None

        port = self.port

        addr_info = self.getaddrinfo(host, port, family=libzt.ZTS_AF_INET, type=libzt.ZTS_SOCK_STREAM)
        if len(addr_info) == 0:
            raise OSError('getaddrinfo returns an empty list')

        while True:
            af, socktype, proto, canonname, sa = addr_info.pop()

            sock: Optional[_socket] = None
            try:
                sock = _socket(af, socktype, proto)

                # If provided, set socket level options before connecting
                if self.socket_options is not None:
                    for opt in self.socket_options:
                        sock.setsockopt(*opt)

                # sock.settimeout(self.timeout)  # TODO - not supported by libzt yet

                if self.source_address is not None:
                    sock.bind(self.source_address)

                sock.connect(sa)
                return sock

            except OSError as e:
                if sock is not None:
                    sock.close()
                if len(addr_info) > 0:
                    # Try another address
                    continue
                else:
                    raise e

    @staticmethod
    def getaddrinfo(host: str, port: int, family: int = 0, type: int = 0, proto: int = 0, flags: int = 0):
        """Return necessary arguments for creating sockets to each specified host
        (can be multiple in comma-separated list)"""
        # TODO
        af = libzt.ZTS_AF_INET
        socktype = libzt.ZTS_SOCK_STREAM
        proto = 0
        canonname = None
        sa = (host, port)

        return [(af, socktype, proto, canonname, sa)]



import urllib3

from requezts import _HTTPConnectionPool


class _PoolManager(urllib3.poolmanager.PoolManager):
    """Requests PoolManager that uses libzt sockets

    Internal object to requezts. Should not be called by user code."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pool_classes_by_scheme = {
            'http': _HTTPConnectionPool,
            # TODO: add HTTPS
        }



import requests

from requezts._poolmanager import _PoolManager


# noinspection PyUnresolvedReferences
class _HTTPAdapter(requests.adapters.HTTPAdapter):
    """requests HTTPAdapter using libzt sockets

    Internal object to requezts. Should not be called by user code.
    """

    def init_poolmanager(self, connections, maxsize, block=..., **pool_kwargs):
        """Initializes a urllib3 PoolManager.
        This method should not be called from user code, and is only
        exposed for use when subclassing the
        :class:`HTTPAdapter <requests.adapters.HTTPAdapter>`.
        :param connections: The number of urllib3 connection pools to cache.
        :param maxsize: The maximum number of connections to save in the pool.
        :param block: Block when no free connections are available.
        :param pool_kwargs: Extra keyword arguments used to initialize the Pool Manager.
        """
        # save these values for pickling
        self._pool_connections = connections
        self._pool_maxsize = maxsize
        self._pool_block = block

        self.poolmanager = _PoolManager(num_pools=connections, maxsize=maxsize, block=block, strict=True,
                                        **pool_kwargs)



import io
import socket
from typing import Optional, Union

import libzt


# noinspection PyPep8Naming
class _socket(libzt.sockets.socket):
    """Wrapper over libzt socket

    Internal object to requezts. Should not be called by user code.
    """

    def __init__(self, sock_family=-1, sock_type=-1, sock_proto=-1, sock_fd=None):
        super().__init__(sock_family, sock_type, sock_proto, sock_fd)

    def has_dualstack_ipv6(self):
        return super().has_dualstack_ipv6()

    @property
    def family(self):
        return super().family()

    @property
    def type(self):
        return super().type()

    @property
    def proto(self):
        return super().proto()

    def create_connection(self, remote_address):
        return super().create_connection(remote_address)

    def create_server(self, local_address, sock_family=libzt.ZTS_AF_INET, backlog=None):
        return super().create_server(local_address, sock_family, backlog)

    def accept(self):
        return super().accept()

    def bind(self, local_address):
        super().bind(local_address)

    def close(self):
        super().close()

    def connect(self, address):
        super().connect(address)

    def connect_ex(self, address):
        return super().connect_ex(address)

    def fileno(self):
        return super().fileno()

    def get_inheritable(self):
        return super().get_inheritable()

    def getblocking(self):
        return super().getblocking()

    def getsockopt(self, level, optname, buflen=None):
        return super().getsockopt(level, optname, buflen)

    def ioctl(self, request, arg=0, mutate_flag=True):
        return super().ioctl(request, arg, mutate_flag)

    def listen(self, backlog=None):
        super().listen(backlog)

    # noinspection PyMethodOverriding
    def makefile(self, mode: Optional[str] = None, buffering: Optional[int] = None, *, encoding: Optional[str] = None,
                 errors: Optional[str] = None, newline: Optional[str] = None) -> io.IOBase:
        """Return an IO object connected to the socket
        Modeled after cpython implementation: https://github.com/python/cpython/blob/main/Lib/socket.py

        :param mode: (optional) File mode (one of ["r", "w", "rw", "rb", "wb", "rwb"], default: "r")
        :param buffering: (optional) Buffer size:
            If positive, will return buffered io object with buffer size = `buffering`.
            If zero, will return unbuffered io object.
            If negative or `None`, will default to io.DEFAULT_BUFFER_SIZE.
        :param encoding: (optional) Text encoding
        :param errors: (optional) Specifies how encoding and decoding errors are handled by io.TextIOWrapper object.
        :param newline: (optional) Specifies how newlines are handled by io.TextIOWrapper object.
        """
        mode = mode or "r"
        if mode not in ["r", "w", "rw", "rb", "wb", "rwb"]:
            raise ValueError(f"Invalid mode: \"{mode}\"")

        if (buffering is None) or (buffering < 0):
            buffering = io.DEFAULT_BUFFER_SIZE

        encoding = encoding or "locale"  # TODO: use io.text_encoding() helper func (available in Python 3.10+)

        reading = "r" in mode
        writing = "w" in mode
        binary = "b" in mode

        if (buffering == 0) and (not binary):
            raise ValueError("Unbuffered streams must be binary")

        raw_mode = ("r" if reading else "") + ("w" if writing else "")  # Remove "b"

        res = _SocketIO(self, raw_mode)

        if buffering > 0:
            # Wrap in buffer object
            if reading and writing:
                res = io.BufferedRWPair(res, res, buffer_size=buffering)
            elif reading:
                res = io.BufferedReader(res, buffer_size=buffering)
            else:
                assert writing
                res = io.BufferedWriter(res, buffer_size=buffering)

            if not binary:
                # Wrap in encoding object
                res = io.TextIOWrapper(res, encoding=encoding, errors=errors, newline=newline)
                # noinspection PyPropertyAccess
                res.mode = mode

        return res

    def recv(self, n_bytes, flags=0):
        return super().recv(n_bytes, flags)

    def recv_into(self, buffer: Union[bytearray, memoryview], n_bytes: int = 0, flags: int = 0) -> int:
        """Read received data into buffer
        Based on cpython implementation: socketmodule.c::sock_recv_into()

        :param buffer: Byte buffer
        :param n_bytes: (optional) Number of bytes to read (default: length of buffer)
        :param flags: (optional) Flags
        :return: Number of bytes read
        """
        if n_bytes < 0:
            raise ValueError("negative buffersize in recv_into")

        n_bytes = n_bytes or len(buffer)

        if len(buffer) < n_bytes:
            raise ValueError("buffer too small for requested bytes")

        if n_bytes > 0:
            # TODO: Bypass recv() and implement this at the native level
            recv_bytes = self.recv(n_bytes, flags=flags)
            n_bytes_recv = len(recv_bytes)
            buffer[:n_bytes_recv] = recv_bytes
            return n_bytes_recv
        else:
            return 0

    def send(self, data, flags=0):
        return super().send(data, flags)

    def sendall(self, data: Union[bytes, bytearray, memoryview], flags: int = 0) -> None:
        """Send all data
        https://stackoverflow.com/questions/34252273/what-is-the-difference-between-socket-send-and-socket-sendall
        """
        while True:
            bytes_sent: int = self.send(data, flags=flags)
            if bytes_sent > 0:
                data = data[bytes_sent:]
                continue
            else:
                break

    def setblocking(self, flag):
        super().setblocking(flag)

    def setsockopt(self, level, optname, value):
        super().setsockopt(level, optname, value)

    def settimeout(self, value):
        # TODO
        pass

    def shutdown(self, how):
        super().shutdown(how)


class _SocketIO(socket.SocketIO):
    """IO implementation on top of libzt sockets

    Internal object to requezts. Should not be called by user code.

    :param sock: requezts Socket
    :param mode: File mode (one of ["r", "w", "rw", "rb", "wb", "rwb"])
    """

    def __init__(self, sock: _socket, mode: str):
        # noinspection PyTypeChecker
        super().__init__(sock, mode)

    def close(self) -> None:
        """Close the SocketIO object"""
        if not self.closed:
            io.RawIOBase.close(self)  # Call super() (bypass socket.SocketIO.close())
            # noinspection PyAttributeOutsideInit
            self._sock = None



import time
from typing import Optional

import requests
from libzt import ZeroTierNode

from requezts import _HTTPAdapter


class Session(requests.Session):
    """|  Context manager class built on ``requests.sessions.Session`` that uses libzt sockets. Manages:
    |  - ZeroTier node state
    |  - Cookie persistence
    |  - Connection pooling
    |  - Configuration

    **Note**: Due to current limitations of libzt, requezts can only manage one ZeroTier network node per process.

    :param net_id: ZeroTier network ID (64-bit unsigned integer, e.g. 0x0123456789abcdef)
    :param node_path: (optional) Path in filesystem of node configuration (will be created if not already existing)
    """

    def __init__(self, net_id: int, node_path: Optional[str] = None):

        # Initialize requests.Session. Will mount default HTTP/HTTPS adapter.
        super().__init__()

        self.__net_id = net_id
        self.__node_path = node_path

        self.__node: Optional[ZeroTierNode] = None

        # noinspection HttpUrlsUsage,PyProtectedMember
        self.mount('http://', _HTTPAdapter())
        # noinspection PyTypeChecker,PyProtectedMember
        self.mount('https://', _HTTPAdapter())  # TODO

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def open(self) -> None:
        """
        Connect to ZeroTier network and bring node online.
        """

        assert not self.is_open(), "session already active"

        # Initialize node
        node = ZeroTierNode()
        if self.__node_path is not None:
            node.init_from_storage(self.__node_path)

        # Bring node online
        node.node_start()
        while not node.node_is_online():
            time.sleep(0.02)

        # Join the net
        node.net_join(self.__net_id)
        while not node.net_transport_is_ready(self.__net_id):
            time.sleep(0.02)

        # TODO: Allow configurable IP address

        self.__node = node

    def close(self) -> None:
        """
        Disconnect from ZeroTier network and bring node offline.
        """

        super().close()  # Call close() on all adapters

        assert self.is_open(), "session already closed"

        node = self.__node
        self.__node = None

        # Bring node down
        node.node_stop()
        node.node_free()

    def is_open(self) -> bool:
        """
        Check if Session is open (connected to ZeroTier network).

        :rtype: bool
        """
        return self.__node is not None
